# reference: https://github.com/pytorch/pytorch/blob/8ea5b572a63b1acc538a9fc8d3862c73739116e8/torch/autograd/functional.py
import jittor as jt

__all__ = ["vjp", "jvp", "jacobian", "hessian", "hvp", "vhp"]

# Utility functions
def _as_tuple_nocheck(x):
    if isinstance(x, tuple):
        return x
    elif isinstance(x, list):
        return tuple(x)
    else:
        return (x,)

def _as_tuple(inp, arg_name=None, fn_name=None):
    # Ensures that inp is a tuple of Tensors
    # Returns whether or not the original inp was a tuple and the tupled version of the input
    if arg_name is None and fn_name is None:
        return _as_tuple_nocheck(inp)

    is_inp_tuple = True
    if not isinstance(inp, tuple):
        inp = (inp,)
        is_inp_tuple = False

    for i, el in enumerate(inp):
        if not isinstance(el, (jt.Var, jt.nn.ComplexNumber)):
            if is_inp_tuple:
                raise TypeError(
                    f"The {arg_name} given to {fn_name} must be either a Tensor or a tuple of Tensors but the"
                    f" value at index {i} has type {type(el)}."
                )
            else:
                raise TypeError(
                    f"The {arg_name} given to {fn_name} must be either a Tensor or a tuple of Tensors but the"
                    f" given {arg_name} has type {type(el)}."
                )

    return is_inp_tuple, inp


def _tuple_postprocess(res, to_unpack):
    # Unpacks a potentially nested tuple of Tensors
    # to_unpack should be a single boolean or a tuple of two booleans.
    # It is used to:
    # - invert _as_tuple when res should match the inp given to _as_tuple
    # - optionally remove nesting of two tuples created by multiple calls to _as_tuple
    if isinstance(to_unpack, tuple):
        assert len(to_unpack) == 2
        if not to_unpack[1]:
            res = tuple(el[0] for el in res)
        if not to_unpack[0]:
            res = res[0]
    else:
        if not to_unpack:
            res = res[0]
    return res


def _grad_preprocess(inputs, create_graph, need_graph):
    # Preprocess the inputs to make sure they require gradient
    # inputs is a tuple of Tensors to preprocess
    # create_graph specifies if the user wants gradients to flow back to the Tensors in inputs
    # need_graph specifies if we internally want gradients to flow back to the Tensors in res
    # Note that we *always* create a new Tensor object to be able to see the difference between
    # inputs given as arguments and the same Tensors automatically captured by the user function.
    # Check this issue for more details on how that can happen: https://github.com/pytorch/pytorch/issues/32576
    res = []
    for inp in inputs:
        if create_graph and inp.requires_grad:
            # Create at least a new Tensor object in a differentiable way
            # Use .reshae() to get a shallow copy
            res.append(inp.reshape(inp.shape))
        else:
            if need_graph:
                ninp = inp.detach().start_grad()
            else:
                ninp = inp.detach().stop_grad()
            res.append(ninp)
    return tuple(res)


def _grad_postprocess(inputs, create_graph):
    # Postprocess the generated Tensors to avoid returning Tensors with history when the user did not
    # request it.
    if isinstance(inputs[0], (jt.Var, jt.nn.ComplexNumber)):
        if not create_graph:
            return tuple(inp.detach() for inp in inputs)
        else:
            return inputs
    else:
        return tuple(_grad_postprocess(inp, create_graph) for inp in inputs)


def _validate_v(v, other, is_other_tuple):
    # This assumes that other is the correct shape, and v should match
    # Both are assumed to be tuples of Tensors
    if len(other) != len(v):
        if is_other_tuple:
            raise RuntimeError(
                f"v is a tuple of invalid length: should be {len(other)} but got {len(v)}."
            )
        else:
            raise RuntimeError("The given v should contain a single Tensor.")

    for idx, (el_v, el_other) in enumerate(zip(v, other)):
        if el_v.shape != el_other.shape:
            prepend = ""
            if is_other_tuple:
                prepend = f"Entry {idx} in "
            raise RuntimeError(
                f"{prepend}v has invalid size: should be {el_other.shape} but got {el_v.shape}."
            )


def _check_requires_grad(inputs, input_type, strict):
    # Used to make all the necessary checks to raise nice errors in strict mode.
    if not strict:
        return

    if input_type not in ["outputs", "grad_inputs", "jacobian", "hessian"]:
        raise RuntimeError("Invalid input_type to _check_requires_grad")
    for i, inp in enumerate(inputs):
        if inp is None:
            # This can only be reached for grad_inputs.
            raise RuntimeError(
                f"The output of the user-provided function is independent of input {i}."
                " This is not allowed in strict mode."
            )
        if not inp.requires_grad:
            if input_type == "hessian":
                raise RuntimeError(
                    f"The hessian of the user-provided function with respect to input {i}"
                    " is independent of the input. This is not allowed in strict mode."
                    " You should ensure that your function is thrice differentiable and that"
                    " the hessian depends on the inputs."
                )
            elif input_type == "jacobian":
                raise RuntimeError(
                    "While computing the hessian, found that the jacobian of the user-provided"
                    f" function with respect to input {i} is independent of the input. This is not"
                    " allowed in strict mode. You should ensure that your function is twice"
                    " differentiable and that the jacobian depends on the inputs (this would be"
                    " violated by a linear function for example)."
                )
            elif input_type == "grad_inputs":
                raise RuntimeError(
                    f"The gradient with respect to input {i} is independent of the inputs of the"
                    " user-provided function. This is not allowed in strict mode."
                )
            else:
                raise RuntimeError(
                    f"Output {i} of the user-provided function does not require gradients."
                    " The outputs must be computed in a differentiable manner from the input"
                    " when running in strict mode."
                )


def _autograd_grad(
    outputs,
    inputs,
    grad_outputs=None,
    create_graph=True,
):
    # Version of grad that accepts `None` in outputs and do not compute gradients for them.
    # This has the extra constraint that inputs has to be a tuple
    assert isinstance(outputs, tuple)
    if grad_outputs is None:
        grad_outputs = (None,) * len(outputs)
    assert isinstance(grad_outputs, tuple)
    assert len(outputs) == len(grad_outputs)

    new_outputs = ()
    new_grad_outputs = ()
    for out, grad_out in zip(outputs, grad_outputs):
        if out is not None and out.requires_grad:
            new_outputs += (out,)
            new_grad_outputs += (grad_out,)

    if len(new_outputs) == 0:
        # No differentiable output, we don't need to call the autograd engine
        return (None,) * len(inputs)
    else:
        acc_loss = None
        for new_output, grad_output in zip(new_outputs, grad_outputs):
            if isinstance(new_output, jt.nn.ComplexNumber):
                if grad_output is not None:
                    loss = (new_output.value * grad_output.value).sum()
                else:
                    loss = new_output.value.sum()
            else:
                if grad_output is not None:
                    new_output = new_output * grad_output
                loss = new_output.sum()
            if acc_loss is None:
                acc_loss = loss
            else:
                acc_loss += loss

        complex_inds = []
        var_inputs = []
        for idx, inp in enumerate(inputs):
            if isinstance(inp, jt.nn.ComplexNumber):
                var_inputs.append(inp.value)
                complex_inds.append(idx)
            else:
                var_inputs.append(inp)
        
        grads = jt.grad(acc_loss, var_inputs, retain_graph=create_graph)
        for complex_ind in complex_inds:
            grads[complex_ind] = jt.nn.ComplexNumber(grads[complex_ind], is_concat_value=True)
        return tuple(grads)


def _fill_in_zeros(grads, refs, strict, create_graph, stage):
    # Used to detect None in the grads and depending on the flags, either replace them
    # with Tensors full of 0s of the appropriate size based on the refs or raise an error.
    # strict and create graph allow us to detect when it is appropriate to raise an error
    # stage gives us information of which backward call we consider to give good error message
    if stage not in ["back", "back_trick", "double_back", "double_back_trick"]:
        raise RuntimeError(f"Invalid stage argument '{stage}' to _fill_in_zeros")

    res = ()
    for i, grads_i in enumerate(grads):
        if grads_i is None:
            if strict:
                if stage == "back":
                    raise RuntimeError(
                        "The output of the user-provided function is independent of "
                        f"input {i}. This is not allowed in strict mode."
                    )
                elif stage == "back_trick":
                    raise RuntimeError(
                        f"The gradient with respect to the input is independent of entry {i}"
                        " in the grad_outputs when using the double backward trick to compute"
                        " forward mode gradients. This is not allowed in strict mode."
                    )
                elif stage == "double_back":
                    raise RuntimeError(
                        "The jacobian of the user-provided function is independent of "
                        f"input {i}. This is not allowed in strict mode."
                    )
                else:
                    raise RuntimeError(
                        "The hessian of the user-provided function is independent of "
                        f"entry {i} in the grad_jacobian. This is not allowed in strict "
                        "mode as it prevents from using the double backward trick to "
                        "replace forward mode AD."
                    )

            refs_i = refs[i]
            if isinstance(refs_i, jt.nn.ComplexNumber):
                grads_i = jt.nn.ComplexNumber(jt.zeros_like(refs_i.value), is_concat_value=True)
            else:
                grads_i = jt.zeros_like(refs_i)
        else:
            if strict and create_graph and not grads_i.requires_grad:
                if "double" not in stage:
                    raise RuntimeError(
                        "The jacobian of the user-provided function is independent of "
                        f"input {i}. This is not allowed in strict mode when create_graph=True."
                    )
                else:
                    raise RuntimeError(
                        "The hessian of the user-provided function is independent of "
                        f"input {i}. This is not allowed in strict mode when create_graph=True."
                    )

        res += (grads_i,)

    return res


# Public API

def vjp(func, inputs, v=None, create_graph=False, strict=False):
    r"""Compute the dot product between a vector ``v`` and the Jacobian of the given function at the point given by the inputs.

    Args:
        func (function): a Python function that takes Tensor inputs and returns
            a tuple of Tensors or a Tensor.
        inputs (tuple of Tensors or Tensor): inputs to the function ``func``.
        v (tuple of Tensors or Tensor): The vector for which the vector
            Jacobian product is computed.  Must be the same size as the output
            of ``func``. This argument is optional when the output of ``func``
            contains a single element and (if it is not provided) will be set
            as a Tensor containing a single ``1``.
        create_graph (bool, optional): If ``True``, both the output and result
            will be computed in a differentiable way. Note that when ``strict``
            is ``False``, the result can not require gradients or be
            disconnected from the inputs.  Defaults to ``False``.
        strict (bool, optional): If ``True``, an error will be raised when we
            detect that there exists an input such that all the outputs are
            independent of it. If ``False``, we return a Tensor of zeros as the
            vjp for said inputs, which is the expected mathematical value.
            Defaults to ``False``.

    Returns:
        output (tuple): tuple with:
            func_output (tuple of Tensors or Tensor): output of ``func(inputs)``

            vjp (tuple of Tensors or Tensor): result of the dot product with
            the same shape as the inputs.
    """
    with jt.enable_grad():
        is_inputs_tuple, inputs = _as_tuple(inputs, "inputs", "vjp")
        inputs = _grad_preprocess(inputs, create_graph=create_graph, need_graph=True)

        outputs = func(*inputs)
        is_outputs_tuple, outputs = _as_tuple(
            outputs, "outputs of the user-provided function", "vjp"
        )
        _check_requires_grad(outputs, "outputs", strict=strict)

        if v is not None:
            _, v = _as_tuple(v, "v", "vjp")
            v = _grad_preprocess(v, create_graph=create_graph, need_graph=False)
            _validate_v(v, outputs, is_outputs_tuple)
        else:
            if len(outputs) != 1 or outputs[0].nelement() != 1:
                raise RuntimeError(
                    "The vector v can only be None if the "
                    "user-provided function returns "
                    "a single Tensor with a single element."
                )

    with jt.enable_grad():
        grad_res = _autograd_grad(outputs, inputs, v, create_graph=create_graph)
        vjp = _fill_in_zeros(grad_res, inputs, strict, create_graph, "back")

    # Cleanup objects and return them to the user
    outputs = _grad_postprocess(outputs, create_graph)
    vjp = _grad_postprocess(vjp, create_graph)

    return _tuple_postprocess(outputs, is_outputs_tuple), _tuple_postprocess(
        vjp, is_inputs_tuple
    )


def jvp(func, inputs, v=None, create_graph=False, strict=False):
    r"""Compute the dot product between the Jacobian of the given function at the point given by the inputs and a vector ``v``.

    Args:
        func (function): a Python function that takes Tensor inputs and returns
            a tuple of Tensors or a Tensor.
        inputs (tuple of Tensors or Tensor): inputs to the function ``func``.
        v (tuple of Tensors or Tensor): The vector for which the Jacobian
            vector product is computed. Must be the same size as the input of
            ``func``. This argument is optional when the input to ``func``
            contains a single element and (if it is not provided) will be set
            as a Tensor containing a single ``1``.
        create_graph (bool, optional): If ``True``, both the output and result
            will be computed in a differentiable way. Note that when ``strict``
            is ``False``, the result can not require gradients or be
            disconnected from the inputs.  Defaults to ``False``.
        strict (bool, optional): If ``True``, an error will be raised when we
            detect that there exists an input such that all the outputs are
            independent of it. If ``False``, we return a Tensor of zeros as the
            jvp for said inputs, which is the expected mathematical value.
            Defaults to ``False``.

    Returns:
        output (tuple): tuple with:
            func_output (tuple of Tensors or Tensor): output of ``func(inputs)``

            jvp (tuple of Tensors or Tensor): result of the dot product with
            the same shape as the output.

    """
    with jt.enable_grad():
        is_inputs_tuple, inputs = _as_tuple(inputs, "inputs", "jvp")
        inputs = _grad_preprocess(inputs, create_graph=create_graph, need_graph=True)

        if v is not None:
            _, v = _as_tuple(v, "v", "jvp")
            v = _grad_preprocess(v, create_graph=create_graph, need_graph=False)
            _validate_v(v, inputs, is_inputs_tuple)
        else:
            if len(inputs) != 1 or inputs[0].nelement() != 1:
                raise RuntimeError(
                    "The vector v can only be None if the input to "
                    "the user-provided function is a single Tensor "
                    "with a single element."
                )

        outputs = func(*inputs)
        is_outputs_tuple, outputs = _as_tuple(
            outputs, "outputs of the user-provided function", "jvp"
        )
        _check_requires_grad(outputs, "outputs", strict=strict)
        # The backward is linear so the value of grad_outputs is not important as
        # it won't appear in the double backward graph. We only need to ensure that
        # it does not contain inf or nan.
        grad_outputs = tuple(
            jt.nn.ComplexNumber(jt.zeros_like(out.value), is_concat_value=True) if isinstance(out, jt.nn.ComplexNumber) else jt.zeros_like(out)
            for out in outputs
        )

        grad_inputs = _autograd_grad(outputs, inputs, grad_outputs=grad_outputs, create_graph=True)
        _check_requires_grad(grad_inputs, "grad_inputs", strict=strict)

    if create_graph:
        with jt.enable_grad():
            grad_res = _autograd_grad(
                grad_inputs, grad_outputs, v, create_graph=create_graph
            )
            jvp = _fill_in_zeros(grad_res, outputs, strict, create_graph, "back_trick")
    else:
        grad_res = _autograd_grad(
            grad_inputs, grad_outputs, v, create_graph=create_graph
        )
        jvp = _fill_in_zeros(grad_res, outputs, strict, create_graph, "back_trick")

    # Cleanup objects and return them to the user
    outputs = _grad_postprocess(outputs, create_graph)
    jvp = _grad_postprocess(jvp, create_graph)

    return _tuple_postprocess(outputs, is_outputs_tuple), _tuple_postprocess(
        jvp, is_outputs_tuple
    )



# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers:
#     Guowei Yang <471184555@qq.com>
#     Dun Liang <randonlang@gmail.com>. 
#
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************

from typing import Optional, Tuple, List
import warnings
import math

import jittor as jt
from jittor import Var
from jittor.nn import Module, Linear, softmax, pad, linear, dropout
from jittor.init import xavier_uniform_, xavier_gauss_, constant_

def _canonical_mask(
        mask: Optional[Var],
        mask_name: str,
        other_type,
        other_name: str,
        target_type,
        check_other: bool = True,
) -> Optional[Var]:

    if mask is not None:
        _mask_dtype = mask.dtype
        _mask_is_float = mask.dtype == jt.float16 or mask.dtype == jt.float32 or mask.dtype == jt.float64
        if _mask_dtype != jt.bool and not _mask_is_float:
            raise AssertionError(
                f"only bool and floating types of {mask_name} are supported")
        if check_other and other_type is not None:
            if _mask_dtype != other_type:
                warnings.warn(
                    f"Support for mismatched {mask_name} and {other_name} "
                    "is deprecated. Use same type for both instead."
                )
        if not _mask_is_float:
            # WARNING(514flowey): Check Here
            new_mask = jt.zeros_like(mask, dtype=target_type)
            new_mask[mask] = float("-inf")
            mask = new_mask
    return mask

def _none_or_dtype(input: Optional[Var]):
    if input is None:
        return None
    elif isinstance(input, jt.Var):
        return input.dtype
    raise RuntimeError("input to _none_or_dtype() must be None or torch.Tensor")

def baddbmm(input_var:jt.Var, batch1:jt.Var, batch2:jt.Var, beta=1, alpha=1) -> jt.Var:
    # WARNING(514flowey): Check here
    return beta * input_var + alpha * (batch1 @ batch2)

def scaled_dot_product_attention(query, key, value, attn_mask=None, dropout_p=0.0, is_causal=False, scale=None) -> jt.Var:
    # Efficient implementation equivalent to the following:
    L, S = query.size(-2), key.size(-2)
    scale_factor = 1 / math.sqrt(query.size(-1)) if scale is None else scale
    attn_bias = jt.zeros(L, S, dtype=query.dtype)
    if is_causal:
        assert attn_mask is None
        temp_mask = jt.ones(L, S, dtype=jt.bool).tril(diagonal=0)
        attn_bias[jt.logical_not(temp_mask)] = float("-inf")
        # attn_bias.to(query.dtype)
        attn_bias = jt.array(attn_bias, query.dtype)

    if attn_mask is not None:
        if attn_mask.dtype == jt.bool:
            attn_bias[jt.logical_not(temp_mask)] = float("-inf")
        else:
            attn_bias += attn_mask
    attn_weight = query @ key.transpose(-2, -1) * scale_factor
    attn_weight += attn_bias
    attn_weight = softmax(attn_weight, dim=-1)
    attn_weight = dropout(attn_weight, dropout_p, train=True)
    return attn_weight @ value

def _mha_shape_check(query: Var, key: Var, value: Var,
                     key_padding_mask: Optional[Var], attn_mask: Optional[Var], num_heads: int):
    if query.dim() == 3:
        is_batched = True
        assert key.dim() == 3 and value.dim() == 3, \
            ("For batched (3-D) `query`, expected `key` and `value` to be 3-D"
             f" but found {key.dim()}-D and {value.dim()}-D tensors respectively")
        if key_padding_mask is not None:
            assert key_padding_mask.dim() == 2, \
                ("For batched (3-D) `query`, expected `key_padding_mask` to be `None` or 2-D"
                 f" but found {key_padding_mask.dim()}-D tensor instead")
        if attn_mask is not None:
            assert attn_mask.dim() in (2, 3), \
                ("For batched (3-D) `query`, expected `attn_mask` to be `None`, 2-D or 3-D"
                 f" but found {attn_mask.dim()}-D tensor instead")
    elif query.dim() == 2:
        is_batched = False
        assert key.dim() == 2 and value.dim() == 2, \
            ("For unbatched (2-D) `query`, expected `key` and `value` to be 2-D"
             f" but found {key.dim()}-D and {value.dim()}-D tensors respectively")

        if key_padding_mask is not None:
            assert key_padding_mask.dim() == 1, \
                ("For unbatched (2-D) `query`, expected `key_padding_mask` to be `None` or 1-D"
                 f" but found {key_padding_mask.dim()}-D tensor instead")

        if attn_mask is not None:
            assert attn_mask.dim() in (2, 3), \
                ("For unbatched (2-D) `query`, expected `attn_mask` to be `None`, 2-D or 3-D"
                 f" but found {attn_mask.dim()}-D tensor instead")
            if attn_mask.dim() == 3:
                expected_shape = (num_heads, query.shape[0], key.shape[0])
                assert attn_mask.shape == expected_shape, \
                    (f"Expected `attn_mask` shape to be {expected_shape} but got {attn_mask.shape}")
    else:
        raise AssertionError(
            f"query should be unbatched 2D or batched 3D tensor but received {query.dim()}-D query tensor")

    return is_batched

def _in_projection_packed(
    q: Var,
    k: Var,
    v: Var,
    w: Var,
    b: Optional[Var] = None,
) -> List[Var]:
    E = q.size(-1)
    if k is v:
        if q is k:
            # self-attention
            proj = linear(q, w, b)
            # reshape to 3, E and not E, 3 is deliberate for better memory coalescing and keeping same order as chunk()
            # proj = proj.unflatten(-1, (3, E)).unsqueeze(0).transpose(0, -2).squeeze(-2).contiguous()
            nshape = proj.shape[:-1] + (3, E)
            proj = proj.reshape(nshape).unsqueeze(0).transpose(0, -2).squeeze(-2)
            return proj[0], proj[1], proj[2]
        else:
            # encoder-decoder attention
            w_q, w_kv = w.split([E, E * 2])
            if b is None:
                b_q = b_kv = None
            else:
                b_q, b_kv = b.split([E, E * 2])
            q_proj = linear(q, w_q, b_q)
            kv_proj = linear(k, w_kv, b_kv)
            # reshape to 2, E and not E, 2 is deliberate for better memory coalescing and keeping same order as chunk()
            # kv_proj = kv_proj.unflatten(-1, (2, E)).unsqueeze(0).transpose(0, -2).squeeze(-2).contiguous()
            nshape = kv_proj.shape[:-1] + (2, E)
            kv_proj = kv_proj.reshape(nshape).unsqueeze(0).transpose(0, -2).squeeze(-2)
            return (q_proj, kv_proj[0], kv_proj[1])
    else:
        w_q, w_k, w_v = w.chunk(3)
        if b is None:
            b_q = b_k = b_v = None
        else:
            b_q, b_k, b_v = b.chunk(3)
        return linear(q, w_q, b_q), linear(k, w_k, b_k), linear(v, w_v, b_v)

def _in_projection(
    q: Var,
    k: Var,
    v: Var,
    w_q: Var,
    w_k: Var,
    w_v: Var,
    b_q: Optional[Var] = None,
    b_k: Optional[Var] = None,
    b_v: Optional[Var] = None,
) -> Tuple[Var, Var, Var]:
    Eq, Ek, Ev = q.size(-1), k.size(-1), v.size(-1)
    assert w_q.shape == (Eq, Eq), f"expecting query weights shape of {(Eq, Eq)}, but got {w_q.shape}"
    assert w_k.shape == (Eq, Ek), f"expecting key weights shape of {(Eq, Ek)}, but got {w_k.shape}"
    assert w_v.shape == (Eq, Ev), f"expecting value weights shape of {(Eq, Ev)}, but got {w_v.shape}"
    assert b_q is None or b_q.shape == (Eq,), f"expecting query bias shape of {(Eq,)}, but got {b_q.shape}"
    assert b_k is None or b_k.shape == (Eq,), f"expecting key bias shape of {(Eq,)}, but got {b_k.shape}"
    assert b_v is None or b_v.shape == (Eq,), f"expecting value bias shape of {(Eq,)}, but got {b_v.shape}"
    return linear(q, w_q, b_q), linear(k, w_k, b_k), linear(v, w_v, b_v)

def multi_head_attention_forward(
    query: Var,
    key: Var,
    value: Var,
    embed_dim_to_check: int,
    num_heads: int,
    in_proj_weight: Optional[Var],
    in_proj_bias: Optional[Var],
    bias_k: Optional[Var],
    bias_v: Optional[Var],
    add_zero_attn: bool,
    dropout_p: float,
    out_proj_weight: Var,
    out_proj_bias: Optional[Var],
    training: bool = True,
    key_padding_mask: Optional[Var] = None,
    need_weights: bool = True,
    attn_mask: Optional[Var] = None,
    use_separate_proj_weight: bool = False,
    q_proj_weight: Optional[Var] = None,
    k_proj_weight: Optional[Var] = None,
    v_proj_weight: Optional[Var] = None,
    static_k: Optional[Var] = None,
    static_v: Optional[Var] = None,
    average_attn_weights: bool = True,
    is_causal: bool = False,
) -> Tuple[Var, Optional[Var]]:

    is_batched = _mha_shape_check(query, key, value, key_padding_mask, attn_mask, num_heads)

    # For unbatched input, we unsqueeze at the expected batch-dim to pretend that the input
    # is batched, run the computation and before returning squeeze the
    # batch dimension so that the output doesn't carry this temporary batch dimension.
    if not is_batched:
        # unsqueeze if the input is unbatched
        query = query.unsqueeze(1)
        key = key.unsqueeze(1)
        value = value.unsqueeze(1)
        if key_padding_mask is not None:
            key_padding_mask = key_padding_mask.unsqueeze(0)

    # set up shape vars
    tgt_len, bsz, embed_dim = query.shape
    src_len, _, _ = key.shape

    key_padding_mask = _canonical_mask(
        mask=key_padding_mask,
        mask_name="key_padding_mask",
        other_type=_none_or_dtype(attn_mask),
        other_name="attn_mask",
        target_type=query.dtype
    )

    if is_causal and attn_mask is None:
        raise RuntimeError(
            "Need attn_mask if specifying the is_causal hint. "
            "You may use the Transformer module method "
            "`generate_square_subsequent_mask` to create this mask."
        )

    if is_causal and key_padding_mask is None and not need_weights:
        # when we have a kpm or need weights, we need attn_mask
        # Otherwise, we use the is_causal hint go as is_causal
        # indicator to SDPA.
        attn_mask = None
    else:
        attn_mask = _canonical_mask(
            mask=attn_mask,
            mask_name="attn_mask",
            other_type=None,
            other_name="",
            target_type=query.dtype,
            check_other=False,
        )

        if key_padding_mask is not None:
            # We have the attn_mask, and use that to merge kpm into it.
            # Turn off use of is_causal hint, as the merged mask is no
            # longer causal.
            is_causal = False

    assert embed_dim == embed_dim_to_check, \
        f"was expecting embedding dimension of {embed_dim_to_check}, but got {embed_dim}"
    if isinstance(embed_dim, jt.Var):
        # embed_dim can be a tensor when JIT tracing
        head_dim = embed_dim.div(num_heads, rounding_mode='trunc')
    else:
        head_dim = embed_dim // num_heads
    assert head_dim * num_heads == embed_dim, f"embed_dim {embed_dim} not divisible by num_heads {num_heads}"
    if use_separate_proj_weight:
        # allow MHA to have different embedding dimensions when separate projection weights are used
        assert key.shape[:2] == value.shape[:2], \
            f"key's sequence and batch dims {key.shape[:2]} do not match value's {value.shape[:2]}"
    else:
        assert key.shape == value.shape, f"key shape {key.shape} does not match value shape {value.shape}"

    #
    # compute in-projection
    #
    if not use_separate_proj_weight:
        assert in_proj_weight is not None, "use_separate_proj_weight is False but in_proj_weight is None"
        q, k, v = _in_projection_packed(query, key, value, in_proj_weight, in_proj_bias)
    else:
        assert q_proj_weight is not None, "use_separate_proj_weight is True but q_proj_weight is None"
        assert k_proj_weight is not None, "use_separate_proj_weight is True but k_proj_weight is None"
        assert v_proj_weight is not None, "use_separate_proj_weight is True but v_proj_weight is None"
        if in_proj_bias is None:
            b_q = b_k = b_v = None
        else:
            b_q, b_k, b_v = in_proj_bias.chunk(3)
        q, k, v = _in_projection(query, key, value, q_proj_weight, k_proj_weight, v_proj_weight, b_q, b_k, b_v)

    # prep attention mask

    if attn_mask is not None:
        # ensure attn_mask's dim is 3
        if attn_mask.dim() == 2:
            correct_2d_size = (tgt_len, src_len)
            if attn_mask.shape != correct_2d_size:
                raise RuntimeError(f"The shape of the 2D attn_mask is {attn_mask.shape}, but should be {correct_2d_size}.")
            attn_mask = attn_mask.unsqueeze(0)
        elif attn_mask.dim() == 3:
            correct_3d_size = (bsz * num_heads, tgt_len, src_len)
            if attn_mask.shape != correct_3d_size:
                raise RuntimeError(f"The shape of the 3D attn_mask is {attn_mask.shape}, but should be {correct_3d_size}.")
        else:
            raise RuntimeError(f"attn_mask's dimension {attn_mask.dim()} is not supported")

    # add bias along batch dimension (currently second)
    if bias_k is not None and bias_v is not None:
        assert static_k is None, "bias cannot be added to static key."
        assert static_v is None, "bias cannot be added to static value."
        k = jt.concat([k, bias_k.repeat(1, bsz, 1)])
        v = jt.concat([v, bias_v.repeat(1, bsz, 1)])
        if attn_mask is not None:
            attn_mask = pad(attn_mask, (0, 1))
        if key_padding_mask is not None:
            key_padding_mask = pad(key_padding_mask, (0, 1))
    else:
        assert bias_k is None
        assert bias_v is None

    #
    # reshape q, k, v for multihead attention and make em batch first
    #
    q = q.view(tgt_len, bsz * num_heads, head_dim).transpose(0, 1)
    if static_k is None:
        k = k.view(k.shape[0], bsz * num_heads, head_dim).transpose(0, 1)
    else:
        # TODO finish disentangling control flow so we don't do in-projections when statics are passed
        assert static_k.size(0) == bsz * num_heads, \
            f"expecting static_k.size(0) of {bsz * num_heads}, but got {static_k.size(0)}"
        assert static_k.size(2) == head_dim, \
            f"expecting static_k.size(2) of {head_dim}, but got {static_k.size(2)}"
        k = static_k
    if static_v is None:
        v = v.view(v.shape[0], bsz * num_heads, head_dim).transpose(0, 1)
    else:
        # TODO finish disentangling control flow so we don't do in-projections when statics are passed
        assert static_v.size(0) == bsz * num_heads, \
            f"expecting static_v.size(0) of {bsz * num_heads}, but got {static_v.size(0)}"
        assert static_v.size(2) == head_dim, \
            f"expecting static_v.size(2) of {head_dim}, but got {static_v.size(2)}"
        v = static_v

    # add zero attention along batch dimension (now first)
    if add_zero_attn:
        zero_attn_shape = (bsz * num_heads, 1, head_dim)
        k = jt.concat([k, jt.zeros(zero_attn_shape, dtype=k.dtype)], dim=1)
        v = jt.concat([v, jt.zeros(zero_attn_shape, dtype=v.dtype)], dim=1)
        if attn_mask is not None:
            attn_mask = pad(attn_mask, (0, 1))
        if key_padding_mask is not None:
            key_padding_mask = pad(key_padding_mask, (0, 1))

    # update source sequence length after adjustments
    src_len = k.size(1)

    # merge key padding and attention masks
    if key_padding_mask is not None:
        assert key_padding_mask.shape == (bsz, src_len), \
            f"expecting key_padding_mask shape of {(bsz, src_len)}, but got {key_padding_mask.shape}"
        key_padding_mask = key_padding_mask.view(bsz, 1, 1, src_len).   \
            expand(-1, num_heads, -1, -1).reshape(bsz * num_heads, 1, src_len)
        if attn_mask is None:
            attn_mask = key_padding_mask
        else:
            attn_mask = attn_mask + key_padding_mask

    # adjust dropout probability
    if not training:
        dropout_p = 0.0

    #
    # (deep breath) calculate attention and out projection
    #

    if need_weights:
        B, Nt, E = q.shape
        q_scaled = q / math.sqrt(E)

        assert not (is_causal and attn_mask is None), "FIXME: is_causal not implemented for need_weights"

        if attn_mask is not None:
            attn_output_weights = baddbmm(attn_mask, q_scaled, k.transpose(-2, -1))
        else:
            attn_output_weights = jt.bmm(q_scaled, k.transpose(-2, -1))
        attn_output_weights = softmax(attn_output_weights, dim=-1)
        if dropout_p > 0.0:
            attn_output_weights = dropout(attn_output_weights, p=dropout_p)

        attn_output = jt.bmm(attn_output_weights, v)

        attn_output = attn_output.transpose(0, 1).contiguous().view(tgt_len * bsz, embed_dim)
        attn_output = linear(attn_output, out_proj_weight, out_proj_bias)
        attn_output = attn_output.view(tgt_len, bsz, attn_output.size(1))

        # optionally average attention weights over heads
        attn_output_weights = attn_output_weights.view(bsz, num_heads, tgt_len, src_len)
        if average_attn_weights:
            attn_output_weights = attn_output_weights.mean(dim=1)

        if not is_batched:
            # squeeze the output if input was unbatched
            attn_output = attn_output.squeeze(1)
            attn_output_weights = attn_output_weights.squeeze(0)
        return attn_output, attn_output_weights
    else:
        # attn_mask can be either (L,S) or (N*num_heads, L, S)
        # if attn_mask's shape is (1, L, S) we need to unsqueeze to (1, 1, L, S)
        # in order to match the input for SDPA of (N, num_heads, L, S)
        if attn_mask is not None:
            if attn_mask.size(0) == 1 and attn_mask.dim() == 3:
                attn_mask = attn_mask.unsqueeze(0)
            else:
                attn_mask = attn_mask.view(bsz, num_heads, -1, src_len)

        q = q.view(bsz, num_heads, tgt_len, head_dim)
        k = k.view(bsz, num_heads, src_len, head_dim)
        v = v.view(bsz, num_heads, src_len, head_dim)

        attn_output = scaled_dot_product_attention(q, k, v, attn_mask, dropout_p, is_causal)
        attn_output = attn_output.permute(2, 0, 1, 3).contiguous().view(bsz * tgt_len, embed_dim)

        attn_output = linear(attn_output, out_proj_weight, out_proj_bias)
        attn_output = attn_output.view(tgt_len, bsz, attn_output.size(1))
        if not is_batched:
            # squeeze the output if input was unbatched
            attn_output = attn_output.squeeze(1)
        return attn_output, None


class MultiheadAttention(Module):
    __constants__ = ['batch_first']
    bias_k: Optional[jt.Var]
    bias_v: Optional[jt.Var]

    def __init__(self, embed_dim, num_heads, dropout=0., bias=True, add_bias_kv=False, add_zero_attn=False,
                 kdim=None, vdim=None, batch_first=False, dtype=jt.float32) -> None:
        if embed_dim <= 0 or num_heads <= 0:
            raise ValueError(
                f"embed_dim and num_heads must be greater than 0,"
                f" got embed_dim={embed_dim} and num_heads={num_heads} instead"
            )
        factory_kwargs = {'dtype': dtype}
        super().__init__()
        self.embed_dim = embed_dim
        self.kdim = kdim if kdim is not None else embed_dim
        self.vdim = vdim if vdim is not None else embed_dim
        self._qkv_same_embed_dim = self.kdim == embed_dim and self.vdim == embed_dim

        self.num_heads = num_heads
        self.dropout = dropout
        self.batch_first = batch_first
        self.head_dim = embed_dim // num_heads
        assert self.head_dim * num_heads == self.embed_dim, "embed_dim must be divisible by num_heads"

        if not self._qkv_same_embed_dim:
            self.q_proj_weight = jt.empty((embed_dim, embed_dim), **factory_kwargs)
            self.k_proj_weight = jt.empty((embed_dim, self.kdim), **factory_kwargs)
            self.v_proj_weight = jt.empty((embed_dim, self.vdim), **factory_kwargs)
            self.in_proj_weight = None
        else:
            self.q_proj_weight = None
            self.k_proj_weight = None
            self.v_proj_weight = None
            self.in_proj_weight = jt.empty((3 * embed_dim, embed_dim), **factory_kwargs)

        if bias:
            self.in_proj_bias = jt.empty(3 * embed_dim, **factory_kwargs)
        else:
            self.in_proj_bias = None
        self.out_proj = Linear(embed_dim, embed_dim, bias=bias)

        if add_bias_kv:
            self.bias_k = jt.empty((1, 1, embed_dim), **factory_kwargs)
            self.bias_v = jt.empty((1, 1, embed_dim), **factory_kwargs)
        else:
            self.bias_k = self.bias_v = None

        self.add_zero_attn = add_zero_attn

        self._reset_parameters()

    def _reset_parameters(self):
        if self._qkv_same_embed_dim:
            xavier_uniform_(self.in_proj_weight)
        else:
            xavier_uniform_(self.q_proj_weight)
            xavier_uniform_(self.k_proj_weight)
            xavier_uniform_(self.v_proj_weight)

        if self.in_proj_bias is not None:
            constant_(self.in_proj_bias, 0.)
            constant_(self.out_proj.bias, 0.)
        if self.bias_k is not None:
            xavier_gauss_(self.bias_k)
        if self.bias_v is not None:
            xavier_gauss_(self.bias_v)

    def __setstate__(self, state):
        # Support loading old MultiheadAttention checkpoints generated by v1.1.0
        if '_qkv_same_embed_dim' not in state:
            state['_qkv_same_embed_dim'] = True

        super().__setstate__(state)

    def execute(
            self,
            query: Var,
            key: Var,
            value: Var,
            key_padding_mask: Optional[Var] = None,
            need_weights: bool = True,
            attn_mask: Optional[Var] = None,
            average_attn_weights: bool = True,
            is_causal : bool = False) -> Tuple[Var, Optional[Var]]:

        #####
        # Fast Path is not Supported.
        #####

        is_batched = query.dim() == 3

        key_padding_mask = _canonical_mask(
            mask=key_padding_mask,
            mask_name="key_padding_mask",
            other_type=_none_or_dtype(attn_mask),
            other_name="attn_mask",
            target_type=query.dtype
        )

        attn_mask = _canonical_mask(
            mask=attn_mask,
            mask_name="attn_mask",
            other_type=None,
            other_name="",
            target_type=query.dtype,
            check_other=False,
        )

        if self.batch_first and is_batched:
            # make sure that the transpose op does not affect the "is" property
            if key is value:
                if query is key:
                    query = key = value = query.transpose(1, 0)
                else:
                    query, key = (x.transpose(1, 0) for x in (query, key))
                    value = key
            else:
                query, key, value = (x.transpose(1, 0) for x in (query, key, value))

        if not self._qkv_same_embed_dim:
            attn_output, attn_output_weights = multi_head_attention_forward(
                query, key, value, self.embed_dim, self.num_heads,
                self.in_proj_weight, self.in_proj_bias,
                self.bias_k, self.bias_v, self.add_zero_attn,
                self.dropout, self.out_proj.weight, self.out_proj.bias,
                training=self.is_training(),
                key_padding_mask=key_padding_mask, need_weights=need_weights,
                attn_mask=attn_mask,
                use_separate_proj_weight=True,
                q_proj_weight=self.q_proj_weight, k_proj_weight=self.k_proj_weight,
                v_proj_weight=self.v_proj_weight,
                average_attn_weights=average_attn_weights,
                is_causal=is_causal)
        else:
            attn_output, attn_output_weights = multi_head_attention_forward(
                query, key, value, self.embed_dim, self.num_heads,
                self.in_proj_weight, self.in_proj_bias,
                self.bias_k, self.bias_v, self.add_zero_attn,
                self.dropout, self.out_proj.weight, self.out_proj.bias,
                training=self.is_training(),
                key_padding_mask=key_padding_mask,
                need_weights=need_weights,
                attn_mask=attn_mask,
                average_attn_weights=average_attn_weights,
                is_causal=is_causal)
        if self.batch_first and is_batched:
            return attn_output.transpose(1, 0), attn_output_weights
        else:
            return attn_output, attn_output_weights



# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved.
# Maintainers: Dun Liang <randonlang@gmail.com>.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import os
from jittor_utils import env_or_try_find
import jittor_utils
import ctypes
import glob
import jittor.compiler as compiler
import jittor as jt

has_acl = 0
cc_flags = ""
tikcc_path = env_or_try_find('tikcc_path', 'ccec')
dlopen_flags = os.RTLD_NOW | os.RTLD_GLOBAL
compiler.has_acl = has_acl

# export LD_LIBRARY_PATH=/usr/local/Ascend/ascend-toolkit/latest/tools/aoe/lib64:/usr/local/Ascend/ascend-toolkit/latest/compiler/lib64:/usr/local/Ascend/ascend-toolkit/latest/compiler/lib64/plugin/opskernel:/usr/local/Ascend/ascend-toolkit/latest/compiler/lib64/plugin/nnengine:/usr/local/Ascend/ascend-toolkit/latest/runtime/lib64:/usr/local/Ascend/ascend-toolkit/latest/compiler/lib64/stub:/usr/local/Ascend/ascend-toolkit/latest/tools/tikicpulib/lib/Ascend910A:/usr/local/Ascend/ascend-toolkit/latest/toolkit/tools/simulator/Ascend910A/lib:/opt/AXESMI/lib64:/usr/local/Ascend/driver/lib64/driver/
# export PYTHONPATH=/home/cjld/new_jittor/jittor/python
# export tikcc_path=g++

# conda activate cann
# source /usr/local/Ascend/ascend-toolkit/set_env.sh
# export PYTHONPATH=/home/cjld/new_jittor/jittor/python:/home/cjld/new_jittor/jittor/my/jtorch/python:$PYTHONPATH
# export TASK_QUEUE_ENABLE=0
# python3 -m jittor.test.test_acl -k array
# jittor: conda activate cann && source /usr/local/Ascend/ascend-toolkit/set_env.sh && PYTHONPATH=/home/cjld/new_jittor/jittor/python:/home/cjld/new_jittor/jittor/my/jtorch/python:$PYTHONPATH && cd /home/cjld/new_jittor/jittor/my/mm_benchmark
# python3 -m jittor.test.test_acl -k test_sum
# export ASCEND_SLOG_PRINT_TO_STDOUT=0
# ASCEND_GLOBAL_LOG_LEVEL
# export DUMP_GE_GRAPH=1
# export DUMP_GRAPH_LEVEL=1

# build pytorch-npu
# bash ./ci/build.sh
# python3 -m pip install ./dist/torch_npu-1.11.0.post1-cp37-cp37m-linux_x86_64.whl  --force-reinstall
# pytorch: conda activate cann && source /usr/local/Ascend/ascend-toolkit/set_env.sh && export TASK_QUEUE_ENABLE=0  && cd /home/cjld/new_jittor/jittor/my/mm_benchmark
# python3 ./mm_bench_pt_npu.py


def install():
    import jittor.compiler as compiler
    global has_acl, cc_flags
    acl_compiler_home = os.path.dirname(__file__)
    cc_files = sorted(glob.glob(acl_compiler_home + "/**/*.cc",
                                recursive=True))
    cc_files2 = []
    for name in cc_files:
        if "acl_op_exec" in name:
            compiler.extra_core_files.append(name)
        else:
            cc_files2.append(name)
    cc_files = cc_files2
    ascend_toolkit_home = os.getenv('ASCEND_TOOLKIT_HOME')
    cc_flags += f" -DHAS_CUDA -DIS_ACL  \
    -I{ascend_toolkit_home}/include/ \
    -L{ascend_toolkit_home}/lib64/ \
    -I{acl_compiler_home} -lascendcl -lacl_op_compiler "

    ctypes.CDLL("libascendcl.so", dlopen_flags)
    f'''
    -ltikc_runtime
    -I/usr/local/Ascend/driver/include/ \
    -L{ascend_toolkit_home}/compiler/lib64/ \
    -L{ascend_toolkit_home}/runtime/lib64/ \
    '''
    jittor_utils.LOG.i("ACL detected")

    global mod
    mod = jittor_utils.compile_module(
        '''
#include "common.h"
namespace jittor {
// @pyjt(process)
string process_acl(const string& src, const string& name, const map<string,string>& kargs);
// @pyjt(init_acl_ops)
void init_acl_ops();
}''', compiler.cc_flags + " " + " ".join(cc_files) + cc_flags)
    jittor_utils.process_jittor_source("acl", mod.process)

    has_acl = 1
    os.environ["use_mkl"] = "0"
    compiler.setup_fake_cuda_lib = True


def install_extern():
    return False


def check():
    import jittor.compiler as compiler
    global has_acl, cc_flags
    if tikcc_path:
        try:
            install()
        except Exception as e:
            jittor_utils.LOG.w(f"load ACL failed, exception: {e}")
            has_acl = 0
    compiler.has_acl = has_acl
    compiler.tikcc_path = tikcc_path
    if not has_acl: return False
    compiler.cc_flags += cc_flags
    compiler.nvcc_path = tikcc_path
    compiler.nvcc_flags = compiler.cc_flags.replace("-std=c++14", "")
    return True


def post_process():
    if has_acl:
        from jittor import pool
        pool.pool_use_code_op = False
        import jittor as jt
        jt.flags.use_cuda_host_allocator = 1
        jt.flags.use_parallel_op_compiler = 0
        jt.flags.amp_reg |= 32 + 4  # 32 keep float16, 4 keep reduce type
        mod.init_acl_ops()


def acl_cmd(name: str, inputs: list, output_dtypes: list, output_shapes: list,
            attr: dict):
    nchw_op = ['MaxPoolWithArgmaxV1', 'MaxPoolGradWithArgmaxV1', 'AvgPoolV2']
    attr_op = [
        'MaxPoolWithArgmaxV1', 'MaxPoolGradWithArgmaxV1', 'AvgPoolV2',
        'AdaptiveAvgPool2d', 'AdaptiveAvgPool2dGrad', 'ReverseV2'
    ]

    input_code = ''
    for i in range(len(inputs)):
        if name in nchw_op:
            input_code += f"op.add(in{i}, true, ACL_FORMAT_NCHW);\n"
        else:
            input_code += f"op.add(in{i}, true);\n"

    output_code = ''
    for i in range(len(output_dtypes)):
        if name in nchw_op:
            output_code += f"op.add(out{i}, false, ACL_FORMAT_NCHW);\n"
        else:
            output_code += f"op.add(out{i}, false);\n"

    # add attr to op
    attr_code = ''
    if name in attr_op:
        for k, v in attr.items():
            if isinstance(v, bool):
                if v == True:
                    attr_code += f"op.set_attr(\"{k}\", 1, 1);\n"
                else:
                    attr_code += f"op.set_attr(\"{k}\", 1, 0);\n"
            elif isinstance(v, str):
                attr_code += f"op.set_attr(\"{k}\", \"{v}\");\n"
            elif k == 'divisor_override_value':
                attr_code += f"op.set_attr(\"{k}\", int64_t({v}), 0);\n"
            else:
                v = str(v).replace('[', '{').replace(']', '}')
                attr_code += f"op.set_attr(\"{k}\", vector<int64_t>{v});\n"
    else:
        for k, v in attr.items():
            if isinstance(v, bool):
                if v == True:
                    attr_code += f"op.set_attr(\"{k}\", 1, 1);\n"
                else:
                    attr_code += f"op.set_attr(\"{k}\", 1, 0);\n"
            elif isinstance(v, str):
                attr_code += f"op.set_attr(\"{k}\", \"{v}\");\n"
            else:
                attr_code += f"op.set_attr(\"{k}\", int({v}));\n"

    #print("input_code",input_code)
    #print("attr_code",attr_code)
    import jittor as jt
    return jt.code(output_shapes,
                   output_dtypes,
                   inputs,
                   cuda_header="""
    #include <acl/acl.h>
    #include <acl/acl_op_compiler.h>
    #include <Python.h>
    #include <pystate.h>

    namespace jittor {
    
    void printDeviceData(const vector<aclTensorDesc*>& output_desc, const vector<aclDataBuffer*>& output_data, const string& name = "", bool input=true) {
        LOGir << "name: " << name;
        if(input)
            LOGir << "is input";
        else
            LOGir << "is ouput";
        for (size_t i = 0; i < output_desc.size(); ++i) {
            void* base_addr = aclGetDataBufferAddr(output_data[i]);
            LOGir << "addr of data[" << i << "] :" << base_addr;
            size_t num_dims = aclGetTensorDescNumDims(output_desc[i]);
            size_t total_size = 1;
            std::vector<int64_t> dims(num_dims);
            
            std::cout << "shape of data: ";
            for (size_t j = 0; j < num_dims; ++j) {
                aclGetTensorDescDimV2(output_desc[i], j, &dims[j]);
                total_size *= dims[j];
                std::cout << dims[j] << ", ";
            }
            int evey_batch_size = total_size/dims[0];
            std::cout << std::endl;

            // for(int i= 0; i < dims[0]; i++) {
            //     evey_batch_size = 16;
            //     std::vector<float> host_buffer(evey_batch_size);
            //     void* offset_addr = static_cast<char*>(base_addr) + i * evey_batch_size * sizeof(float);
            //     aclrtMemcpy(host_buffer.data(), evey_batch_size * sizeof(float), offset_addr, evey_batch_size * sizeof(float), ACL_MEMCPY_DEVICE_TO_HOST);
            //     std::cout << "batch[" << i << "]:";
            //     for (size_t k = 0; k < evey_batch_size; ++k) {
            //         std::cout << host_buffer[k] << ", ";
            //     }
            //     std::cout << std::endl;
            // }
        }
    }

    struct AclOpRunner {
        string name;
        vector<aclTensorDesc*> input_desc;
        vector<aclTensorDesc*> output_desc;
        vector<aclDataBuffer*> input_data;
        vector<aclDataBuffer*> output_data;
        aclopAttr *attr;
        vector<vector<uint64>> input_host;
        vector<vector<int>> input_host_32;

        AclOpRunner(const string& name) : name(name) {
            attr = aclopCreateAttr();
        }

        ~AclOpRunner() {
            for (auto i : input_desc) aclDestroyTensorDesc(i);
            for (auto i : output_desc) aclDestroyTensorDesc(i);
            for (auto i : input_data) aclDestroyDataBuffer(i);
            for (auto i : output_data) aclDestroyDataBuffer(i);
            aclopDestroyAttr(attr);
        }

        aclDataType get_dtype(NanoString s) {
            if (s == ns_float32) return ACL_FLOAT;
            if (s == ns_float16) return ACL_FLOAT16;
            if (s == ns_int64) return ACL_INT64;
            if (s == ns_int32) return ACL_INT32;
            if (s == ns_int8) return ACL_INT8;
            if (s == ns_int16) return ACL_INT16;
            if (s == ns_uint8) return ACL_UINT8;
            if (s == ns_uint16) return ACL_UINT16;
            if (s == ns_uint32) return ACL_UINT32;
            if (s == ns_bool) return ACL_BOOL;
            LOGf << "Not supported dtype: " << s;
            return ACL_FLOAT;
        }

        void add(Var* v, bool is_input, int format=ACL_FORMAT_ND) {
            int64_t shape[v->shape.size()];
            for (int i=0; i<v->shape.size(); i++) shape[i] = v->shape[i];

            auto desc = aclCreateTensorDesc(get_dtype(v->dtype()), v->shape.size(), &shape[0], (aclFormat)format);
            aclSetTensorFormat(desc, (aclFormat)format);
            aclSetTensorShape(desc, v->shape.size(), &shape[0]);
            LOGv << "aclCreateTensorDesc" << (int)get_dtype(v->dtype()) << v->shape.size() << &shape[0] << format;
            auto data = aclCreateDataBuffer(v->mem_ptr, v->size);
            LOGv << "aclCreateDataBuffer" << v->mem_ptr << v->size;
            ASSERT(desc && data);
            if (is_input) {
                input_desc.push_back(desc);
                input_data.push_back(data);
            } else {
                output_desc.push_back(desc);
                output_data.push_back(data);
            }
        }

        void add_input_host(vector<uint64> v, int dtype=ACL_UINT64) {
            int64_t shape[1];
            shape[0] = v.size();
            auto desc = aclCreateTensorDesc((aclDataType)dtype, 1, &shape[0], ACL_FORMAT_ND);
            aclSetTensorPlaceMent(desc, ACL_MEMTYPE_HOST_COMPILE_INDEPENDENT);
            LOGv << "aclCreateTensorDesc" << dtype << 1 << &shape[0] << ACL_FORMAT_ND;
            auto data = aclCreateDataBuffer(&v[0], v.size()*sizeof(uint64));
            ASSERT(desc && data);
            LOGv << "aclCreateDataBuffer" << &v[0] << v.size()*sizeof(uint64);
            input_desc.push_back(desc);
            input_data.push_back(data);
            input_host.emplace_back(move(v));
            LOGv << "move" << input_host.back().data();
        }

        void add_input_host_scalar(vector<uint64> v, int dtype=ACL_UINT32) {
            int64_t shape[1];
            shape[0] = v.size();
            auto x = (int*)&v[0];
            x[0] = (int32)v[0];
            auto desc = aclCreateTensorDesc((aclDataType)dtype, 0, &shape[0], ACL_FORMAT_ND);
            aclSetTensorPlaceMent(desc, ACL_MEMTYPE_HOST_COMPILE_INDEPENDENT);
            LOGv << "aclCreateTensorDesc" << dtype << 1 << &shape[0] << ACL_FORMAT_ND;
            auto data = aclCreateDataBuffer(&v[0], v.size()*sizeof(uint32));
            ASSERT(desc && data);
            LOGv << "aclCreateDataBuffer" << &v[0] << v.size()*sizeof(uint32);
            input_desc.push_back(desc);
            input_data.push_back(data);
            input_host.emplace_back(move(v));
        }

        void add_input_host_nv(NanoVector nv, int dtype=ACL_UINT64) {
            vector<uint64> v(nv.size());
            for (int i=0; i<nv.size(); i++) v[i] = nv[i];
            int64_t shape[1];
            shape[0] = v.size();
            auto desc = aclCreateTensorDesc((aclDataType)dtype, 1, &shape[0], ACL_FORMAT_ND);
            aclSetTensorPlaceMent(desc, ACL_MEMTYPE_HOST_COMPILE_INDEPENDENT);
            auto data = aclCreateDataBuffer(&v[0], v.size()*sizeof(uint64));
            input_desc.push_back(desc);
            input_data.push_back(data);
            input_host.emplace_back(move(v));
        }

        void add_input_host_nv32(NanoVector nv, int dtype=ACL_INT32) {
            vector<int> v(nv.size());
            for (int i=0; i<nv.size(); i++) v[i] = nv[i];
            int64_t shape[1];
            shape[0] = v.size();
            auto desc = aclCreateTensorDesc((aclDataType)dtype, 1, &shape[0], ACL_FORMAT_ND);
            LOGv << "aclCreateTensorDesc" << dtype << 1 << &shape[0] << ACL_FORMAT_ND << v;
            aclSetTensorPlaceMent(desc, ACL_MEMTYPE_HOST);
            auto data = aclCreateDataBuffer(&v[0], v.size()*sizeof(int));
            input_desc.push_back(desc);
            input_data.push_back(data);
            input_host_32.emplace_back(move(v));
        }
        
        void set_attr(const string& key, bool value) {
            // LOGir << "string bool" << "set_attr" << key << value;
            CHECK(aclopSetAttrBool(attr, key.c_str(), value)==0);
        }
        void set_attr(const string& key, int value, int is_bool) {
            // LOGir << "string bool" << "set_attr" << key << value << is_bool;
            CHECK(aclopSetAttrBool(attr, key.c_str(), value==is_bool)==0);
        }
        void set_attr(const string& key, float value) {
            // LOGir << "string float" <<"set_attr" << key << value;
            CHECK(aclopSetAttrFloat(attr, key.c_str(), value)==0);
        }
        void set_attr(const string& key, int64_t value) {
            // LOGir << "string int64" << "set_attr" << key << value;
            CHECK(aclopSetAttrInt(attr, key.c_str(), value)==0);
        }
        void set_attr(const string& key, int64_t value, int placeholder) {
            // LOGir << "string int64" << "set_attr" << key << value;
            CHECK(aclopSetAttrInt(attr, key.c_str(), value)==0);
        }
        void set_attr(const string& key, int32 value) {
            // LOGir << "string int32" << "set_attr" << key << value;
            CHECK(aclopSetAttrInt(attr, key.c_str(), value)==0);
        }
        void set_attr(const string& key, vector<int64_t> value) {
            // LOGir << "string vector" << "set_attr" << key << value;
            CHECK(aclopSetAttrListInt(attr, key.c_str(), value.size(), &value[0])==0);
        }
        void set_attr(const string& key, string value) {
            // LOGir << "string string" << "set_attr" << key << value;
            CHECK(aclopSetAttrString(attr, key.c_str(), value.c_str())==0);
        }
        void set_attr(const char* key, const char* value) {
            // LOGir << "char" << "set_attr" << key << value;
            CHECK(aclopSetAttrString(attr, key, value)==0);
        }

        void run() {
            // printDeviceData(input_desc, input_data, name);
            
            LOGv << "run" << name << input_desc.size() << output_desc.size();
            if (!PyGILState_Check()) {
                ASSERT(0==aclopCompileAndExecuteV2(name.c_str(), input_desc.size(), &input_desc[0], &input_data[0], output_desc.size(), &output_desc[0], &output_data[0], attr, ACL_ENGINE_SYS, ACL_COMPILE_SYS, NULL, aclstream));
            } else {
                int ret;
                Py_BEGIN_ALLOW_THREADS
                ret = aclopCompileAndExecuteV2(name.c_str(), input_desc.size(), &input_desc[0], &input_data[0], output_desc.size(), &output_desc[0], &output_data[0], attr, ACL_ENGINE_SYS, ACL_COMPILE_SYS, NULL, aclstream);
                Py_END_ALLOW_THREADS
                if (ret != 0)
                    LOGf << "aclopCompileAndExecuteV2" << name << "failed return" << ret;
            }
            ASSERT(0==aclrtSynchronizeDevice());

            // printDeviceData(output_desc, output_data, name, false);
        }
    };

    }
    """,
                   cuda_src=f"""
    // aclop
    AclOpRunner op("{name}");
    {input_code}
    {output_code}
    {attr_code}
    op.run();""")


def change_function():
    import jittor as jt
    from jittor import Function

    class IndexACL(Function):

        def __init__(self):
            super(IndexACL, self).__init__()

        def execute(self, inshape: list, dim, dtype="int32"):
            # zeros a tensor, shape is inshape, dtype is dtype
            dim_input = dim
            if dim == None:
                dim = [i for i in range(len(inshape))]
            elif type(dim) == int:
                dim = [dim]
            results = []
            for d in dim:
                max_len = inshape[d]
                tmp = jt.zeros(max_len, dtype=dtype)
                result = acl_cmd(
                    "Range", [jt.Var(0), jt.Var(max_len),
                              jt.Var(1)],
                    output_dtypes=[tmp.dtype],
                    output_shapes=[tmp.shape],
                    attr={})[0]
                broadcast_dim = []
                for i in range(len(inshape)):
                    if i != d:
                        broadcast_dim.append(i)
                result = jt.broadcast(result,
                                      shape=inshape,
                                      dims=broadcast_dim)
                results.append(result)
            if len(results) != 1 or dim_input == None:
                return tuple(results)
            else:
                return results[0]

        def grad(self, grad_output):
            return grad_output

    class PoolACL(Function):

        def get_paddings(self):
            pad_top = self.padding[0]
            pad_left = self.padding[1]
            H = self.input.shape[-2]
            W = self.input.shape[-1]

            totalH = H + 2 * self.padding[0] - self.kernel_size[0]
            totalW = W + 2 * self.padding[1] - self.kernel_size[1]

            kH = (totalH + self.stride[0] -
                  1) // self.stride[0] + 1 if self.attr[
                      'ceil_mode'] else totalH // self.stride[0] + 1
            kW = (totalW + self.stride[1] -
                  1) // self.stride[1] + 1 if self.attr[
                      'ceil_mode'] else totalW // self.stride[1] + 1

            if self.attr['ceil_mode']:
                if (kH - 1) * self.stride[0] >= H + self.padding[0]:
                    kH -= 1
                    need_pad_h = (kH -
                                  1) * self.stride[0] + self.kernel_size[0] - H
                    pad_top = need_pad_h - self.padding[0]
                if (kW - 1) * self.stride[1] >= W + self.padding[1]:
                    kW -= 1
                    need_pad_w = (kW -
                                  1) * self.stride[1] + self.kernel_size[1] - W
                    pad_left = need_pad_w - self.padding[1]

            pads = [self.padding[0], pad_top, self.padding[1], pad_left]
            return pads

        def __init__(self,
                     kernel_size,
                     stride=None,
                     padding=0,
                     dilation=None,
                     return_indices=None,
                     ceil_mode=False,
                     count_include_pad=True,
                     op='maximum'):
            super(PoolACL, self).__init__()
            # set attr
            self.kernel_size = kernel_size if isinstance(
                kernel_size, tuple) else (kernel_size, kernel_size)
            stride = stride if stride else kernel_size
            self.stride = stride if isinstance(stride, tuple) else (stride,
                                                                    stride)
            self.padding = padding if isinstance(padding, tuple) else (padding,
                                                                       padding)
            dilation = dilation if dilation else 1
            self.dilation = dilation if isinstance(
                dilation, tuple) else (dilation, dilation)
            attr = {}

            self.return_indices = return_indices
            self.uint16 = jt.Var(1).int32().dtype
            self.op = op

            if op == 'mean':
                attr['exclusive'] = not count_include_pad
                attr['global_pooling'] = False
                attr['divisor_override_value'] = 0
                attr['ksize'] = [
                    1, 1, self.kernel_size[0], self.kernel_size[1]
                ]
                attr['strides'] = [1, 1, self.stride[0], self.stride[1]]
                attr['ceil_mode'] = ceil_mode
                attr['padding_mode'] = 'CALCULATED'
                attr['data_format'] = 'NCHW'
            elif op == 'maximum':
                attr['ksize'] = [
                    1, self.kernel_size[0], self.kernel_size[1], 1
                ]
                attr['strides'] = [1, self.stride[0], self.stride[1], 1]
                attr['pads'] = [1, self.padding[0], self.padding[1], 1]
                attr['dilation'] = [1, self.dilation[0], self.dilation[1], 1]
                # attr['ceil_mode'] = ceil_mode

            self.attr = attr

        def execute(self, input):

            # create input
            input_shape = input.shape
            input_dtype = input.dtype

            self.input = input
            # create output
            output_shape = [
                input_shape[0], input_shape[1],
                (input_shape[2] + 2 * self.padding[0] - self.dilation[0] *
                 (self.kernel_size[0] - 1) - 1) // self.stride[0] + 1,
                (input_shape[3] + 2 * self.padding[1] - self.dilation[1] *
                 (self.kernel_size[1] - 1) - 1) // self.stride[1] + 1
            ]
            output_dtype = input_dtype

            if self.op == 'mean':
                self.attr['pads'] = self.get_paddings()
                result = acl_cmd("AvgPoolV2", [input],
                                 output_dtypes=[output_dtype],
                                 output_shapes=[output_shape],
                                 attr=self.attr)
            elif self.op == 'maximum':
                result = acl_cmd("MaxPoolWithArgmaxV1", [input],
                                 output_dtypes=[output_dtype, self.uint16],
                                 output_shapes=[output_shape, output_shape],
                                 attr=self.attr)
            else:
                raise ValueError('no this type pool')

            if self.op == 'maximum':
                self.index = result[1]

            if self.return_indices:
                return result[0], result[1]
            else:
                return result[0]

        def grad(self, grad_output):
            if self.op == 'maximum':
                grad_input = acl_cmd("MaxPoolGradWithArgmaxV1",
                                     [self.input, grad_output, self.index],
                                     output_dtypes=[grad_output.dtype],
                                     output_shapes=[self.input.shape],
                                     attr=self.attr)[0]
            elif self.op == 'mean':
                grad_input = acl_cmd("AvgPoolV2",
                                     [self.input, grad_output, self.index],
                                     output_dtypes=[grad_output.dtype],
                                     output_shapes=[self.input.shape],
                                     attr=self.attr)[0]
            else:
                grad_input = None
            return grad_input

    class BmmACL(Function):

        def __init__(self, adj_x1=False, adj_x2=False):
            super(BmmACL, self).__init__()
            self.adj_x1 = adj_x1
            self.adj_x2 = adj_x2

        def execute(self, x1, x2):
            self.input = [x1, x2]
            result = acl_cmd("BatchMatMul", [x1, x2],
                             output_dtypes=[x1.dtype],
                             output_shapes=[x1.shape[:-1] + x2.shape[-1:]],
                             attr={})[0]
            return result

        def grad(self, grad_output):
            x1, x2 = self.input
            grad_x1 = acl_cmd(
                "BatchMatMul", [grad_output, x2.transpose(-2, -1)],
                output_dtypes=[x1.dtype],
                output_shapes=[grad_output.shape[:-1] + x1.shape[-1:]],
                attr={})[0]
            grad_x2 = acl_cmd(
                "BatchMatMul", [x1.transpose(-2, -1), grad_output],
                output_dtypes=[x2.dtype],
                output_shapes=[x2.shape[:-1] + grad_output.shape[-1:]],
                attr={})[0]
            return grad_x1, grad_x2

    class MatmulACL(Function):

        def __init__(self, adj_x1=False, adj_x2=False):
            super(MatmulACL, self).__init__()
            self.adj_x1 = adj_x1
            self.adj_x2 = adj_x2

        def execute(self, x1, x2):
            self.input = [x1, x2]
            if len(x1.shape) > 2 or len(x2.shape) > 2:
                result = acl_cmd("BatchMatMul", [x1, x2],
                                 output_dtypes=[x1.dtype],
                                 output_shapes=[x1.shape[:-1] + x2.shape[-1:]],
                                 attr={})[0]
            else:
                result = acl_cmd("MatMul", [x1, x2],
                                 output_dtypes=[x1.dtype],
                                 output_shapes=[x1.shape[:-1] + x2.shape[-1:]],
                                 attr={})[0]
            return result

        def grad(self, grad_output):
            x1, x2 = self.input
            if len(x1.shape) > 2 or len(x2.shape) > 2:
                grad_x1 = acl_cmd(
                    "BatchMatMul",
                    [grad_output, x2.transpose(-2, -1)],
                    output_dtypes=[x1.dtype],
                    output_shapes=[grad_output.shape[:-1] + x1.shape[-1:]],
                    attr={})[0]
                grad_x2 = acl_cmd(
                    "BatchMatMul", [x1.transpose(-2, -1), grad_output],
                    output_dtypes=[x2.dtype],
                    output_shapes=[x2.shape[:-1] + grad_output.shape[-1:]],
                    attr={})[0]
            else:
                grad_x1 = acl_cmd(
                    "MatMul", [grad_output, x2.transpose(-2, -1)],
                    output_dtypes=[x1.dtype],
                    output_shapes=[grad_output.shape[:-1] + x1.shape[-1:]],
                    attr={})[0]
                grad_x2 = acl_cmd(
                    "MatMul", [x1.transpose(-2, -1), grad_output],
                    output_dtypes=[x2.dtype],
                    output_shapes=[x2.shape[:-1] + grad_output.shape[-1:]],
                    attr={})[0]
            return grad_x1, grad_x2

    class GetItem(Function):

        def __init__(self):
            super(GetItem, self).__init__()
            self.type_ = 'index'

        def stride(self, x, dim):
            stride = 1
            for i in range(dim + 1, len(x.shape)):
                stride *= x.shape[i]
            return stride

        def execute(self, x, slices, return_x=None):
            if isinstance(slices, jt.Var) or isinstance(slices, tuple):
                if isinstance(slices, jt.Var):
                    slices = (slices, )
                if isinstance(slices[0], jt.Var):
                    slices_len = len(slices)
                    masks = jt.ones(slices_len, dtype=jt.int64)
                    output = slices[0].shape
                    output += x.shape[slices_len:]
                    input_ = [x, masks, jt.Var(list(output)).int64()]
                    for i in range(slices_len):
                        input_.append(slices[i].int32())
                    result = acl_cmd("Index",
                                     input_,
                                     output_dtypes=[x.dtype],
                                     output_shapes=[output],
                                     attr={})[0]
                    self.shape = x.shape
                    self.sizes = list(output)
                    self.type_ = 'index'
                    self.slices = slices
                    # self.strides
                    return result

            # use AsStrided operator to implement the getitem function
            # get the shape and stride of the input tensor
            x_dim = len(x.shape)
            # int type
            if not isinstance(slices, tuple):
                slices = (slices, )

            if len(slices) < x_dim:
                slices += (slice(None, None, None), ) * (x_dim - len(slices))

            self.inputs = [x, slices]

            sizes = []
            strides = []
            offset = 0

            for dim, s in enumerate(slices):
                if isinstance(s, int):
                    if s < 0:  # Handle negative indices.
                        s += x.shape[dim]
                    offset += s * self.stride(x, dim)
                elif isinstance(s, slice):
                    # Unpack the slice
                    start, stop, step = s.indices(x.size(dim))
                    size = (stop - start - 1) // step + 1
                    stride = self.stride(x, dim) * step
                    offset += start * self.stride(x, dim)
                    sizes.append(size)
                    strides.append(stride)
                else:
                    raise ValueError("Invalid slice type")

            if not sizes:
                sizes = [1]
                strides = [0]
            # AsStrided same with as_strided of pytorch
            self.sizes = sizes
            self.strides = strides
            self.offset = offset
            self.shape = x.shape
            self.type_ = 'as_strided'
            result = acl_cmd(
                "AsStrided",
                [x, jt.Var(sizes),
                 jt.Var(strides),
                 jt.Var(offset)],
                output_dtypes=[x.dtype],
                output_shapes=[jt.empty(sizes).shape],
                attr={})[0]
            return result

        def grad(self, grad_output):
            if self.type_ == 'as_strided':
                result = jt.zeros(self.shape, dtype=grad_output.dtype)
                sizes = list(grad_output.shape)
                strides = [
                    self.stride(grad_output, dim)
                    for dim in range(len(grad_output.shape))
                ]
                result = acl_cmd("ViewCopy", [
                    result,
                    jt.Var(self.sizes),
                    jt.Var(self.strides),
                    jt.Var(self.offset), grad_output,
                    jt.Var(sizes),
                    jt.Var(strides),
                    jt.Var(0)
                ],
                                 output_dtypes=[result.dtype],
                                 output_shapes=[result.shape],
                                 attr={})[0]
            elif self.type_ == 'index':
                #TODO: use IndexPutV2 to implement the grad function
                assert len(self.slices) == 1
                index = self.slices[0]
                input = jt.zeros(self.shape, dtype=grad_output.dtype)
                input_flatten = input.reshape(input.shape[0], -1)
                index_flatten = index.reshape(-1).unsqueeze(-1).repeat(
                    1, input_flatten.shape[1])
                grad_output_flatten = grad_output.reshape(index.numel(), -1)
                result = acl_cmd(
                    "ScatterElements",
                    [input_flatten, index_flatten, grad_output_flatten],
                    output_dtypes=[input.dtype],
                    output_shapes=[input.shape],
                    attr={
                        'axis': 0,
                        'reduction': 'add'
                    })[0]
                result = result.reshape(self.shape)
                # result = jt.zeros(self.shape, dtype=grad_output.dtype)
                # # masks = jt.ones(len(self.slices), dtype=jt.int64)
                # masks = jt.array([1,1], dtype=jt.int64)
                # expand_masks = jt.array([1,1], dtype=jt.int64)
                # inputs_ = [result,grad_output,masks,expand_masks]
                # slices_len = len(self.slices)
                # for  i in range(slices_len):
                #     inputs_.append(self.slices[i].int64())
                # # breakpoint()
                # jt.sync_all(True)
                # print(inputs_)
                # result_ = acl_cmd("IndexPutV2", inputs_,
                #                  output_dtypes=[result.dtype],
                #                  output_shapes=[result.shape],
                #                  attr={"accumulate":True})[0]
                # result = result_
            else:
                raise ValueError("Invalid slice type")
            result.sync()
            return result, None

    class ConcatACL(Function):

        def __init__(self):
            super(ConcatACL, self).__init__()

        def execute(self, input_tensors, dim=0):
            self.input = input_tensors
            for i in range(len(input_tensors)):
                if input_tensors[i].dtype != input_tensors[0].dtype:
                    raise ValueError(
                        "All input tensors must have the same dtype")
                if input_tensors[i].shape[:dim] != input_tensors[
                        0].shape[:dim] or input_tensors[i].shape[
                            dim + 1:] != input_tensors[0].shape[dim + 1:]:
                    raise ValueError(
                        "All input tensors must have the same shape")
            result = acl_cmd(
                "ConcatD",
                input_tensors,
                output_dtypes=[input_tensors[0].dtype],
                output_shapes=[
                    jt.empty(self.calculate_output_shape(input_tensors,
                                                         dim)).shape
                ],
                attr={
                    "N": len(input_tensors),
                    "concat_dim": dim
                })[0]
            return result

        def grad(self, grad_output):
            grad_inputs = self.split_grad(grad_output, self.input, self.axis)
            return grad_inputs

        def calculate_output_shape(self, input_tensors, axis):
            shape = list(input_tensors[0].shape)
            for tensor in input_tensors[1:]:
                shape[axis] += tensor.shape[axis]
            return tuple(shape)

        def split_grad(self, grad_output, input_tensors, axis):
            offset = 0
            grad_inputs = []
            for tensor in input_tensors:
                grad_input = acl_cmd("Slice", [
                    grad_output, [0] * axis + [offset] + [0] *
                    (len(tensor.shape) - axis - 1), tensor.shape
                ])
                grad_inputs.append(grad_input)
                offset += tensor.shape[axis]
            return grad_inputs

    class SetItemACL(Function):

        def __init__(self):
            super(SetItemACL, self).__init__()

        def stride(self, x, dim):
            # 计算给定维度的步长
            stride = 1
            for i in range(dim + 1, len(x.shape)):
                stride *= x.shape[i]
            return stride

        def execute(self, x, slices, value, reduce='void'):
            self.is_tensor = type(value) == jt.Var
            if type(value) != jt.Var:
                value = jt.array(value)
            x_dim = len(x.shape)

            # 确保slices是一个元组
            if not isinstance(slices, tuple):
                slices = (slices, )

            # 补齐slices使其长度等于x的维度
            if len(slices) < x_dim:
                slices += (slice(None, None, None), ) * (x_dim - len(slices))

            self.inputs = [x, slices, value]

            target_sizes = []
            target_strides = []
            offset = 0

            for dim, s in enumerate(slices):
                if isinstance(s, int):
                    if s < 0:
                        s += x.shape[dim]
                    s = slice(s, s + 1, None)
                if isinstance(s, slice):
                    # 解包切片
                    start, stop, step = s.indices(x.shape[dim])
                    size = (stop - start - 1) // step + 1
                    stride = self.stride(x, dim) * step
                    offset += start * self.stride(x, dim)
                    target_sizes.append(size)
                    target_strides.append(stride)
                else:
                    print("slices: ", s, type(s))
                    raise ValueError("Invalid slice type")

            # 计算value的size、stride和offset
            value_sizes = list(value.shape)
            value_strides = [
                self.stride(value, dim) for dim in range(len(value.shape))
            ]

            self.target_sizes = target_sizes
            self.target_strides = target_strides
            self.offset = offset
            self.value_sizes = value_sizes
            self.value_strides = value_strides

            #import pdb; pdb.set_trace()
            result = acl_cmd("ViewCopy", [
                x,
                jt.Var(target_sizes),
                jt.Var(target_strides),
                jt.Var(offset), value,
                jt.Var(value_sizes),
                jt.Var(value_strides),
                jt.Var(0)
            ],
                             output_dtypes=[x.dtype],
                             output_shapes=[x.shape],
                             attr={})[0]
            result.sync()
            return result

        def grad(self, grad_output):
            result = acl_cmd("AsStrided", [
                grad_output,
                jt.Var(self.target_sizes),
                jt.Var(self.target_strides),
                jt.Var(self.offset)
            ],
                             output_dtypes=[grad_output.dtype],
                             output_shapes=[jt.empty(self.target_sizes).shape],
                             attr={})[0]
            # copy grad_output to new_grad_output
            new_grad_output = acl_cmd("Copy", [grad_output],
                                      output_dtypes=[grad_output.dtype],
                                      output_shapes=[grad_output.shape],
                                      attr={"N": 1})[0]
            new_grad_output = acl_cmd("ViewCopy", [
                new_grad_output,
                jt.Var(self.target_sizes),
                jt.Var(self.target_strides),
                jt.Var(self.offset),
                jt.zeros(self.value_sizes, dtype=grad_output.dtype),
                jt.Var(self.value_sizes),
                jt.Var(self.value_strides),
                jt.Var(0)
            ],
                                      output_dtypes=[grad_output.dtype],
                                      output_shapes=[grad_output.shape],
                                      attr={})[0]
            new_grad_output.sync()
            return new_grad_output, None, result if self.is_tensor else None

    class TriuACL(Function):

        def __init__(self):
            super(TriuACL, self).__init__()

        def execute(self, input, k):
            self.input = input
            result = acl_cmd("Triu", [input],
                             output_dtypes=[input.dtype],
                             output_shapes=[input.shape],
                             attr={'diagonal': k})[0]
            return result

        def grad(self, grad_output):
            return grad_output

    class TransposeACL(Function):

        def __init__(self):
            super(TransposeACL, self).__init__()

        def execute(self, input, perm):
            self.input = input

            output_shape = input.shape[perm[0]:perm[0] + 1]

            for i in range(1, len(perm)):
                output_shape += input.shape[perm[i]:perm[i] + 1]
            result = acl_cmd("Transpose", [input, jt.Var(perm)],
                             output_dtypes=[input.dtype],
                             output_shapes=[output_shape],
                             attr={})[0]
            return result

        def grad(self, grad_output):
            return grad_output

    class AdaptiveMaxPool2dACL(Function):

        def __init__(
            self,
            output_size,
            return_indices=False,
        ):
            super(AdaptiveMaxPool2dACL, self).__init__()
            self.output_size = (output_size, output_size) if isinstance(
                output_size, int) else output_size

            self.return_indices = return_indices
            self.uint16 = jt.Var(1).int32().dtype

            attr = {}
            attr['ceil_mode'] = False
            attr['dilations'] = [1, 1, 1, 1]
            self.attr = attr

        def execute(self, input):
            input_shape = input.shape
            input_dtype = input.dtype

            output_shape = [
                input_shape[0], input_shape[1], self.output_size[0],
                self.output_size[1]
            ]
            output_dtype = input_dtype
            self.input = input

            stride_h = input_shape[2] // output_shape[2]
            stride_w = input_shape[3] // output_shape[3]
            kernel_size_h = input_shape[2] - (output_shape[2] - 1) * stride_h
            kernel_size_w = input_shape[3] - (output_shape[3] - 1) * stride_w

            stride = [0, 0]
            kernel_size = [0, 0]
            padding = [0, 0]

            stride[0] = stride_h
            stride[1] = stride_w
            kernel_size[0] = kernel_size_h
            kernel_size[1] = kernel_size_w
            padding[0] = padding[1] = 0
            kernel_sizes = [1, kernel_size[0], kernel_size[1], 1]
            strides_size = [1, stride[0], stride[1], 1]
            paddings = [1, padding[0], padding[1], 1]

            self.attr['ksize'] = kernel_sizes
            self.attr['strides'] = strides_size
            self.attr['pads'] = paddings

            result = acl_cmd("MaxPoolWithArgmaxV1", [input],
                             output_dtypes=[output_dtype, self.uint16],
                             output_shapes=[output_shape, output_shape],
                             attr=self.attr)

            self.index = result[1]

            if self.return_indices:
                return result[0], result[1]
            else:
                return result[0]

        def grad(self, grad_output):
            grad_input = acl_cmd("MaxPoolGradWithArgmaxV1",
                                 [self.input, grad_output, self.index],
                                 output_dtypes=[grad_output.dtype],
                                 output_shapes=[self.input.shape],
                                 attr=self.attr)[0]
            return grad_input

    class AdaptiveAvgPool2dACL(Function):

        def __init__(self, output_size):
            super(AdaptiveAvgPool2dACL, self).__init__()
            self.output_size = (output_size, output_size) if isinstance(
                output_size, int) else output_size

            attr = {}
            if isinstance(output_size, tuple):
                output_size = [output_size[0], output_size[1]]
            attr['output_size'] = output_size
            self.attr = attr

        def execute(self, input):
            input_shape = input.shape
            input_dtype = input.dtype

            self.original_shape = input_shape

            output_shape = [
                input_shape[0], input_shape[1], self.attr['output_size'][0],
                self.attr['output_size'][1]
            ]
            output_dtype = input_dtype
            self.input = input

            result = acl_cmd("AdaptiveAvgPool2d", [input],
                             output_dtypes=[output_dtype],
                             output_shapes=[output_shape],
                             attr=self.attr)

            return result[0]

        def grad(self, grad_output):
            attr = {}
            attr['orig_input_shape'] = list(self.original_shape)
            grad_input = acl_cmd("AdaptiveAvgPool2dGrad", [grad_output],
                                 output_dtypes=[grad_output.dtype],
                                 output_shapes=[self.original_shape],
                                 attr=attr)[0]
            return grad_input

    class CumsumACL(Function):

        def __init__(self):
            super(CumsumACL, self).__init__()

        def execute(self, input, dim=-1):
            self.input = input
            self.dim = dim
            result = acl_cmd("Cumsum", [input, jt.Var(dim)],
                             output_dtypes=[input.dtype],
                             output_shapes=[input.shape],
                             attr={})[0]
            return result

        def grad(self, grad_output):
            flipped_grad_output = acl_cmd(
                "ReverseV2", [grad_output, jt.Var([self.dim])],
                output_dtypes=[grad_output.dtype],
                output_shapes=[grad_output.shape],
                attr={})[0]
            cumulative_grad = acl_cmd(
                "Cumsum",
                [flipped_grad_output, jt.Var(self.dim)],
                output_dtypes=[grad_output.dtype],
                output_shapes=[grad_output.shape],
                attr={})[0]
            grad_input = acl_cmd(
                "ReverseV2",
                [cumulative_grad, jt.Var([self.dim])],
                output_dtypes=[grad_output.dtype],
                output_shapes=[grad_output.shape],
                attr={})[0]
            return grad_input

    class GatherACL(Function):

        def __init__(self):
            super(GatherACL, self).__init__()

        def execute(self, input, dim, index):
            self.input = input
            self.dim = dim
            self.index = index

            result = acl_cmd("GatherElements", [input, index],
                             output_dtypes=[input.dtype],
                             output_shapes=[index.shape],
                             attr={'dim': dim})[0]
            return result

        def grad(self, grad_output):
            tmp = jt.zeros(self.index.shape, dtype=grad_output.dtype)
            grad_input = acl_cmd("ScatterElements",
                                 [tmp, self.index, grad_output],
                                 output_dtypes=[grad_output.dtype],
                                 output_shapes=[tmp.shape],
                                 attr={
                                     'axis': self.dim,
                                     'reduction': "add"
                                 })[0]
            return grad_input

    class ScatterACL(Function):

        def __init__(self):
            super(ScatterACL, self).__init__()

        def execute(self, input, dim, index, src, reduce='void'):
            self.input = input
            self.dim = dim
            self.index = index
            self.reduce = reduce
            result = acl_cmd("ScatterElements", [input, self.index, src],
                             output_dtypes=[input.dtype],
                             output_shapes=[input.shape],
                             attr={
                                 'axis': self.dim,
                                 'reduction': reduce
                             })[0]
            return result

        def grad(self, grad_output):
            grad_input = acl_cmd("GatherElements", [grad_output, self.index],
                                 output_dtypes=[grad_output.dtype],
                                 output_shapes=[self.index.shape],
                                 attr={'dim': self.dim})[0]
            return grad_output, None, None, grad_input

    class WhereACL(Function):

        def __init__(self):
            super(WhereACL, self).__init__()

        def execute(self, condition, x, y):
            self.condition = condition

            if x.dtype != y.dtype:
                if x.dtype == jt.float32:
                    y = y.float32()
                elif y.dtype == jt.float32:
                    x = x.float32()
                else:
                    x = x.to(y.dtype)

            self.x = x
            self.y = y

            result = acl_cmd("Select", [condition, x, y],
                             output_dtypes=[x.dtype],
                             output_shapes=[x.shape],
                             attr={})[0]
            return result

        def grad(self, grad_output):
            tmp = jt.zeros(grad_output.shape, dtype=grad_output.dtype)
            grad_x = acl_cmd("Select", [self.condition, grad_output, tmp],
                             output_dtypes=[self.x.dtype],
                             output_shapes=[self.x.shape],
                             attr={})[0]

            grad_y = acl_cmd("Select", [self.condition, tmp, grad_output],
                             output_dtypes=[self.y.dtype],
                             output_shapes=[self.y.shape],
                             attr={})[0]
            return grad_output, grad_x, grad_y

    class FlipACL(Function):

        def __init__(self):
            super(FlipACL, self).__init__()

        def execute(self, input, dim):
            self.input = input
            #if isinstance(dim_vector, tuple):
            dim_vector = jt.Var(list(dim))
            #print(dim_vector.dtype)
            self.dim_vector = dim_vector
            #print(input, dim_vector)
            result = acl_cmd("ReverseV2", [input, dim_vector],
                             output_dtypes=[input.dtype],
                             output_shapes=[input.shape],
                             attr={})[0]
            return result

        def grad(self, grad_output):
            #print(grad_output)
            grad_input = acl_cmd("ReverseV2", [grad_output, self.dim_vector],
                                 output_dtypes=[grad_output.dtype],
                                 output_shapes=[grad_output.shape],
                                 attr={})[0]
            return grad_input

    class FloorIntACL(Function):

        def __init__(self):
            super(FloorIntACL, self).__init__()

        def execute(self, input):
            self.input = input
            self.shape = input.shape
            result = acl_cmd("Floor", [input],
                             output_dtypes=[jt.int],
                             output_shapes=[input.shape],
                             attr={})[0]
            return result

        def grad(self, grad_output):
            return jt.zeros(self.shape, dtype=grad_output.dtype)

    def warp(origin_func, new_func):

        def warpper(*args, **kwargs):
            if origin_func == jt.index:
                if len(args) == 2 and args[1] == None:
                    args = tuple(list(args[0:1]))
            if jt.flags.use_acl:
                if isinstance(new_func, IndexACL):
                    if len(args) == 1:
                        args = (args[0], None)
                if isinstance(new_func, CumsumACL):
                    args = (args[0], kwargs.get('dim', -1))
                    kwargs = {}
                if isinstance(new_func,
                              ScatterACL) and kwargs.get('reduce') is not None:
                    args = (args[0], args[1], args[2], args[3],
                            kwargs.get('reduce', 'void'))
                    kwargs = {}

                return new_func(*args, **kwargs)
            return origin_func(*args, **kwargs)

        return warpper

    jt.index = warp(jt.index, IndexACL())
    jt.Var.index = lambda x, dim=None: warp(jt.index, IndexACL())(x.shape, dim)
    jt.nn.Pool = warp(jt.nn.Pool, PoolACL)
    jt.nn.AdaptiveMaxPool2d = warp(jt.nn.AdaptiveMaxPool2d,
                                   AdaptiveMaxPool2dACL)
    jt.nn.AdaptiveAvgPool2d = warp(jt.nn.AdaptiveAvgPool2d,
                                   AdaptiveAvgPool2dACL)

    jt.triu = warp(jt.triu, TriuACL())
    jt.triu_ = warp(jt.triu, TriuACL())
    jt.Var.triu = lambda x: warp(jt.Var.triu, TriuACL())(x)
    jt.Var.triu_ = lambda x: warp(jt.Var.triu_, TriuACL())(x)

    jt.getitem = warp(jt.getitem, GetItem())
    jt.Var.getitem = lambda x, slices, return_x=None: warp(
        jt.getitem, GetItem())(x, slices)

    jt.setitem = warp(jt.setitem, SetItemACL())
    jt.Var.setitem = lambda x, slices, value, reduce='void': warp(
        jt.setitem, SetItemACL())(x, slices, value, reduce)

    jt.misc.flip = warp(jt.misc.flip, FlipACL())
    jt.Var.flip = lambda x, dim_vector: warp(jt.misc.flip, FlipACL())(
        x, dim_vector)
    jt.cumsum = warp(jt.cumsum, CumsumACL())
    jt.gather = warp(jt.gather, GatherACL())
    jt.Var.gather = lambda x, dim, index: warp(jt.gather, GatherACL())(x, dim,
                                                                       index)
    jt.scatter = warp(jt.scatter, ScatterACL())
    jt.Var.scatter = lambda x, dim, index, src, reduce="void": warp(
        jt.scatter, ScatterACL())(x, dim, index, src, reduce)
    jt.where = warp(jt.where, WhereACL())
    jt.floor_int = warp(jt.floor_int, FloorIntACL())
    jt.Var.floor_int = lambda x: warp(jt.floor_int, FloorIntACL())(x)

    # jt.nn.bmm = warp(jt.nn.bmm, BmmACL())
    # jt.bmm = warp(jt.bmm, BmmACL())
    # jt.nn.matmul = warp(jt.matmul, MatmulACL())
    # jt.matmul = warp(jt.matmul, MatmulACL())
    # jt.transpose = warp(jt.transpose, TransposeACL())
    # jt.Var.transpose = lambda x, perm: warp(jt.transpose, TransposeACL())(x, perm)
    # jt.concat = warp(jt.concat, ConcatACL())



import jittor as jt
from jittor import nn
import io
import pickle
import sys
import torch

class HalfStorage: pass
class BFloat16Storage: pass
class FloatStorage: pass
class LongStorage: pass
class IntStorage: pass
class ShortStorage: pass
class CharStorage: pass
class BoolStorage: pass
HalfStorage.__module__ = "torch"
BFloat16Storage.__module__ = "torch"
FloatStorage.__module__ = "torch"
LongStorage.__module__ = "torch"
IntStorage.__module__ = "torch"
ShortStorage.__module__ = "torch"
CharStorage.__module__ = "torch"
BoolStorage.__module__ = "torch"
def _rebuild_tensor_v2(*args): pass
_rebuild_tensor_v2.__module__ = "torch._utils"

targets = [HalfStorage, BFloat16Storage, FloatStorage, LongStorage, IntStorage, ShortStorage, CharStorage, BoolStorage, _rebuild_tensor_v2]

def swap_targets(targets):
    original_targets = []
    for target in targets:
        original_targets.append(sys.modules[target.__module__].__dict__.get(target.__name__, target))
        sys.modules[target.__module__].__dict__[target.__name__] = target
    return original_targets

class TensorStorage:
    def __init__(self, data):
        self.data = data

class TensorWrapper:
    def __init__(self, data):
        self.data = data
    def __reduce__(self):
        a = tuple(self.data.shape)
        # calc stride
        stride = [1]
        for i in range(len(a)-1, 0, -1):
            stride.append(stride[-1]*a[i])
        stride = stride[::-1]

        return (_rebuild_tensor_v2, (
            TensorStorage(self.data),
            0,
            tuple(self.data.shape),
            tuple(stride),
            False,
            {}
        ))

dtype_map = {
    "float16": HalfStorage,
    "bfloat16": BFloat16Storage,
    "float32": FloatStorage,
    "int64": LongStorage,
    "int32": IntStorage,
    "int16": ShortStorage,
    "int8": CharStorage,
    "bool": BoolStorage,
}

def save_pytorch(path, obj):

    serialized_storages = []
    # dfs and wrap jt.Var into TensorWrapper
    def dfs(obj):
        if isinstance(obj, jt.Var):
            return TensorWrapper(obj)
        elif isinstance(obj, dict):
            return {k: dfs(v) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [dfs(x) for x in obj]
        elif isinstance(obj, tuple):
            return tuple(dfs(x) for x in obj)
        else:
            return obj
    
    def persistent_id(obj):
        if isinstance(obj, TensorStorage):
            storage_type = dtype_map[str(obj.data.dtype)]
            storage_key = len(serialized_storages)
            serialized_storages.append(obj.data)
            storage_numel = obj.data.numel()
            location = 'cpu'
            return ('storage',
                    storage_type,
                    storage_key,
                    location,
                    storage_numel)
        return None

    obj = dfs(obj)
    data_buf = io.BytesIO()
    pickle_protocol = 2
    pickler = pickle.Pickler(data_buf, protocol=pickle_protocol)
    pickler.persistent_id = persistent_id
    global targets
    targets = swap_targets(targets)
    pickler.dump(obj)
    targets = swap_targets(targets)
    data_value = data_buf.getvalue()
    
    # use previous pytorch code to save data
    # from torch.serialization import _open_zipfile_writer
    # with _open_zipfile_writer(path) as zip_file:
    #     print(data_value)
    #     zip_file.write_record('data.pkl', data_value, len(data_value))
    #     zip_file.write_record('byteorder', sys.byteorder, len(sys.byteorder))
    #     for i, v in enumerate(serialized_storages):
    #         b = v.numpy().tobytes()
    #         zip_file.write_record(f'data/{i}', b, len(b))

    import os
    path_base_name = os.path.basename(path).split(".")[0]
    contents = jt.ZipFile(path, "w")
    def write(name, data):
        if isinstance(data, str):
            write(name, data.encode())
        elif isinstance(data, bytes):
            import ctypes
            pointer = ctypes.cast(data, ctypes.c_void_p).value
            contents.write(path_base_name+'/'+name, pointer, len(data))
        elif isinstance(data, jt.Var):
            contents.write(path_base_name+'/'+name, data.raw_ptr, data.nbytes)
        else:
            raise TypeError(f"unsupported type {type(data)}")
    write("data.pkl", data_value)
    write("byteorder", sys.byteorder)
    for i, v in enumerate(serialized_storages):
        write(f"data/{i}", v)
    write("version", "3")
    del contents


if __name__ == "__main__":
    linear = nn.Linear(3, 3)
    save_pytorch("linear.bin", linear.state_dict())

    import torch
    res = torch.load("linear.bin")
    print(res)
    print(linear.state_dict())